"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generatePersistedQueryManifest = exports.getFilepaths = exports.defaults = void 0;
const cache_1 = require("@apollo/client/cache");
const core_1 = require("@apollo/client/core");
const persisted_query_lists_1 = require("@apollo/persisted-query-lists");
const graphql_tag_pluck_1 = require("@graphql-tools/graphql-tag-pluck");
const globby_1 = __importDefault(require("globby"));
const graphql_1 = require("graphql");
const lodash_1 = require("lodash");
const node_crypto_1 = require("node:crypto");
const node_fs_1 = require("node:fs");
const node_path_1 = require("node:path");
const vfile_1 = __importDefault(require("vfile"));
const vfile_reporter_1 = __importDefault(require("vfile-reporter"));
const chalk_1 = __importDefault(require("chalk"));
exports.defaults = {
    documents: [
        "src/**/*.{graphql,gql,js,jsx,ts,tsx}",
        "!**/*.d.ts",
        "!**/*.spec.{js,jsx,ts,tsx}",
        "!**/*.story.{js,jsx,ts,tsx}",
        "!**/*.test.{js,jsx,ts,tsx}",
    ],
    output: "persisted-query-manifest.json",
    createOperationId: (query) => {
        return (0, node_crypto_1.createHash)("sha256").update(query).digest("hex");
    },
};
const COLORS = {
    identifier: chalk_1.default.magenta,
    filepath: chalk_1.default.underline.cyan,
    name: chalk_1.default.yellow,
};
const ERROR_MESSAGES = {
    anonymousOperation: (node) => {
        return `Anonymous GraphQL operations are not supported. Please name your ${node.operation}.`;
    },
    uniqueFragment: (name, source) => {
        return `Fragment named "${COLORS.name(name)}" already defined in: ${COLORS.filepath(source.file.path)}`;
    },
    uniqueOperation: (name, source) => {
        return `Operation named "${COLORS.name(name)}" already defined in: ${COLORS.filepath(source.file.path)}`;
    },
    uniqueOperationId: (id, operationName, definedOperationName) => {
        return `\`createOperationId\` created an ID (${COLORS.identifier(id)}) for operation named "${COLORS.name(operationName)}" that has already been used for operation named "${COLORS.name(definedOperationName)}".`;
    },
    parseError(error) {
        return `${error.name}: ${error.message}`;
    },
};
function addError(source, message) {
    const vfileMessage = source.file.message(message, source.location);
    vfileMessage.fatal = true;
}
function parseLocationFromError(error) {
    if (error instanceof graphql_1.GraphQLError && error.locations) {
        return error.locations[0];
    }
    const loc = "loc" in error &&
        typeof error.loc === "object" &&
        error.loc !== null &&
        error.loc;
    const line = loc && "line" in loc && typeof loc.line === "number" && loc.line;
    const column = loc && "column" in loc && typeof loc.column === "number" && loc.column;
    if (typeof line === "number" && typeof column === "number") {
        return { line, column };
    }
    return;
}
function getDocumentSources(filepath) {
    const file = (0, vfile_1.default)({
        path: filepath,
        contents: (0, node_fs_1.readFileSync)(filepath, "utf-8"),
    });
    try {
        if (file.extname === ".graphql" || file.extname === ".gql") {
            return [
                {
                    node: (0, graphql_1.parse)(file.toString()),
                    file,
                    location: { line: 1, column: 1 },
                },
            ];
        }
        return (0, graphql_tag_pluck_1.gqlPluckFromCodeStringSync)(filepath, file.toString()).map((source) => ({
            node: (0, graphql_1.parse)(source.body),
            file,
            location: source.locationOffset,
        }));
    }
    catch (e) {
        const error = e;
        const source = {
            node: null,
            file,
            location: parseLocationFromError(error),
        };
        addError(source, ERROR_MESSAGES.parseError(error));
        return [source];
    }
}
function maybeReportErrorsAndExit(files) {
    if (!Array.isArray(files)) {
        files = [files];
    }
    if (files.some((file) => file.messages.length > 0)) {
        console.error((0, vfile_reporter_1.default)(files, { quiet: true }));
        process.exit(1);
    }
}
function uniq(arr) {
    return [...new Set(arr)];
}
async function getFilepaths(documents) {
    return [...uniq(await (0, globby_1.default)(documents))].sort((a, b) => a.localeCompare(b));
}
exports.getFilepaths = getFilepaths;
async function generatePersistedQueryManifest(config = {}, configFilePath) {
    const { documents = exports.defaults.documents, createOperationId = exports.defaults.createOperationId, } = config;
    const configFile = (0, vfile_1.default)({
        path: configFilePath
            ? (0, node_path_1.relative)(process.cwd(), configFilePath)
            : "<virtual>",
    });
    const filepaths = await getFilepaths(documents);
    const sources = filepaths.flatMap(getDocumentSources);
    const fragmentsByName = new Map();
    const operationsByName = new Map();
    for (const source of sources) {
        if (!source.node) {
            continue;
        }
        (0, graphql_1.visit)(source.node, {
            FragmentDefinition(node) {
                const name = node.name.value;
                const sources = fragmentsByName.get(name) ?? [];
                if (sources.length) {
                    sources.forEach((sibling) => {
                        addError(source, ERROR_MESSAGES.uniqueFragment(name, sibling));
                        addError(sibling, ERROR_MESSAGES.uniqueFragment(name, source));
                    });
                }
                fragmentsByName.set(name, [...sources, source]);
                return false;
            },
            OperationDefinition(node) {
                const name = node.name?.value;
                if (!name) {
                    addError(source, ERROR_MESSAGES.anonymousOperation(node));
                    return false;
                }
                const sources = operationsByName.get(name) ?? [];
                if (sources.length) {
                    sources.forEach((sibling) => {
                        addError(source, ERROR_MESSAGES.uniqueOperation(name, sibling));
                        addError(sibling, ERROR_MESSAGES.uniqueOperation(name, source));
                    });
                }
                operationsByName.set(name, [...sources, source]);
                return false;
            },
        });
    }
    maybeReportErrorsAndExit(uniq(sources.map((source) => source.file)));
    const fragments = (0, cache_1.createFragmentRegistry)(...sources.map(({ node }) => node).filter(Boolean));
    const manifestOperationIds = new Map();
    const manifestOperations = [];
    const client = new core_1.ApolloClient({
        cache: new core_1.InMemoryCache({
            fragments,
        }),
        link: new core_1.ApolloLink((operation) => {
            const body = (0, graphql_1.print)((0, persisted_query_lists_1.sortTopLevelDefinitions)(operation.query));
            const name = operation.operationName;
            const type = operation.query.definitions.find((d) => d.kind === "OperationDefinition").operation;
            const id = createOperationId(body, {
                operationName: name,
                type,
                createDefaultId() {
                    return exports.defaults.createOperationId(body);
                },
            });
            if (manifestOperationIds.has(id)) {
                addError({ file: configFile }, ERROR_MESSAGES.uniqueOperationId(id, name, manifestOperationIds.get(id)));
            }
            else {
                manifestOperationIds.set(id, name);
            }
            manifestOperations.push({ id, name, type, body });
            return core_1.Observable.of({ data: null });
        }),
    });
    for (const [_, sources] of (0, lodash_1.sortBy)([...operationsByName.entries()], lodash_1.first)) {
        for (const source of sources) {
            if (source.node) {
                await client.query({ query: source.node, fetchPolicy: "no-cache" });
            }
        }
    }
    maybeReportErrorsAndExit(configFile);
    return {
        format: "apollo-persisted-query-manifest",
        version: 1,
        operations: manifestOperations,
    };
}
exports.generatePersistedQueryManifest = generatePersistedQueryManifest;
//# sourceMappingURL=index.js.map